@@ -6,7 +6,7 @@
 ;; Keywords: help
 ;; URL: https://github.com/astoff/devdocs.el
 ;; Package-Requires: ((emacs "27.1"))
-;; Version: 0.3
+;; Version: 0.4
 
 ;; This program is free software; you can redistribute it and/or modify
 ;; it under the terms of the GNU General Public License as published by
@@ -75,6 +75,16 @@ directory-local variable."
   "String used to format a documentation location, e.g. in header line."
   :type 'string)
 
+(defcustom devdocs-disambiguated-entry-format #("%s (%s)" 3 7 (face italic))
+  "How to disambiguate entries with identical names in `devdocs-lookup'.
+This string is passed to `format' with two arguments, the entry
+name and a count."
+  :type '(choice (const :tag "Count in parentheses, italicized"
+                        #("%s (%s)" 3 7 (face italic)))
+                 (const :tag "Invisible cookie"
+                        #("%s (%s)" 2 7 (invisible t)))
+                 string))
+
 (defcustom devdocs-fontify-code-blocks t
   "Whether to fontify code snippets inside pre tags.
 Fontification is done using the `org-src' library, which see."
@@ -107,7 +117,7 @@ its return value; take the necessary precautions."
          (timer-set-time (car data) (time-add nil devdocs-cache-timeout)))
      (let ((val (funcall fun))
            (timer (run-at-time devdocs-cache-timeout nil
-                               (lambda () (remhash funrep devdocs--cache)))))
+                               #'remhash funrep devdocs--cache)))
        (prog1 val
          (puthash funrep (cons timer val) devdocs--cache)))))
 
@@ -159,11 +169,11 @@ otherwise, offer only installed documents.
 
 Return a document metadata alist if MULTIPLE is nil; otherwise, a
 list of metadata alists."
-  (let ((cands (seq-map (lambda (it) (cons (alist-get 'slug it) it))
-                        (if available
-                            (devdocs--available-docs)
-                          (or (devdocs--installed-docs)
-                              (user-error "No documents in `%s'" devdocs-data-dir))))))
+  (let ((cands (mapcar (lambda (it) (cons (alist-get 'slug it) it))
+                       (if available
+                           (devdocs--available-docs)
+                         (or (devdocs--installed-docs)
+                             (user-error "No documents in `%s'" devdocs-data-dir))))))
     (if multiple
         (delq nil (mapcar (lambda (s) (cdr (assoc s cands)))
                           (completing-read-multiple prompt cands)))
@@ -192,8 +202,8 @@ DOC is a document metadata alist."
          pages)
     (with-temp-buffer
       (url-insert-file-contents (format "%s/%s/db.json?%s" devdocs-cdn-url slug mtime))
-      (seq-doseq (entry (let ((json-key-type 'string))
-                          (json-read)))
+      (dolist (entry (let ((json-key-type 'string))
+                       (json-read)))
         (with-temp-file (expand-file-name
                          (url-hexify-string (format "%s.html" (car entry))) temp)
           (push (car entry) pages)
@@ -201,8 +211,8 @@ DOC is a document metadata alist."
     (with-temp-buffer
       (url-insert-file-contents (format "%s/%s/index.json?%s" devdocs-cdn-url slug mtime))
       (let ((index (json-read)))
+        (push `(pages . ,(vconcat (nreverse pages))) index)
         (with-temp-file (expand-file-name "index" temp)
-          (push `(pages . ,(apply #'vector (nreverse pages))) index)
           (prin1 index (current-buffer)))))
     (with-temp-file (expand-file-name "metadata" temp)
       (prin1 (cons devdocs--data-format-version doc) (current-buffer)))
@@ -236,22 +246,20 @@ DOC is a document metadata alist."
 
 ;;; Document indexes
 
-(defun devdocs--index (doc)
-  "Return the index of document DOC.
-This is an alist containing `entries', `pages' and `types'."
-  (let* ((docid (cons 'doc doc))
-         (idx (with-temp-buffer
-                (insert-file-contents (expand-file-name
-                                       (concat (alist-get 'slug doc) "/index")
-                                       devdocs-data-dir))
-                (read (current-buffer))))
-         (entries (alist-get 'entries idx)))
-    (prog1 idx
-      (seq-do-indexed (lambda (entry i)
-                        (push docid entry)
-                        (push `(index . ,i) entry)
-                        (aset entries i entry))
-                      entries))))
+(defun devdocs--index (doc kind)
+  "Return an index of document DOC, where KIND is `entries', `pages' or `types'."
+  (if kind
+      (alist-get kind (devdocs--with-cache (devdocs--index doc nil)))
+    (let* ((docmeta (cons 'doc doc))
+           (indexes (with-temp-buffer
+                      (insert-file-contents (expand-file-name
+                                             (concat (alist-get 'slug doc) "/index")
+                                             devdocs-data-dir))
+                      (read (current-buffer))))
+           (entries (alist-get 'entries indexes)))
+      (prog1 indexes
+        (seq-do-indexed (lambda (entry i) (aset entries i (cons docmeta entry)))
+                        entries)))))
 
 ;;; Documentation viewer
 
@@ -265,7 +273,7 @@ This is an alist containing `entries', `pages' and `types'."
   '(:eval (let-alist (car devdocs--stack)
             (concat (devdocs--doc-title .doc)
                     (and .type devdocs-separator) .type
-                    (and .name devdocs-separator) .name))))
+                    devdocs-separator (or .name .path)))))
 
 (define-derived-mode devdocs-mode special-mode "DevDocs"
   "Major mode for viewing DevDocs documents."
@@ -280,7 +288,8 @@ This is an alist containing `entries', `pages' and `types'."
   "Go to the original position in a DevDocs buffer."
   (interactive)
   (goto-char (point-min))
-  (when-let ((match (text-property-search-forward 'shr-target-id shr-target-id t)))
+  (when-let ((pred (if (fboundp 'shr--set-target-ids) #'member t)) ;; shr change in Emacs 29
+             (match (text-property-search-forward 'shr-target-id shr-target-id pred)))
     (goto-char (prop-match-beginning match))))
 
 (defun devdocs-go-back ()
@@ -305,28 +314,52 @@ Note that this refers to the index order, which may not coincide
 with the order of appearance in the text."
   (interactive "p")
   (let-alist (car devdocs--stack)
-    (devdocs--render
-     (or (ignore-error 'args-out-of-range
-           (seq-elt (alist-get 'entries (devdocs--index .doc))
-                    (+ count .index)))
-         (user-error (if (< count 0) "No previous entry" "No next entry"))))))
+    (let* ((entries (devdocs--index .doc 'entries))
+           (pred (lambda (entry _) (string= (alist-get 'path entry) .path)))
+           (current (seq-position entries nil pred)))
+      (unless current (user-error "No current entry"))
+      (devdocs--render
+       (or (ignore-error 'args-out-of-range (elt entries (+ count current)))
+           (user-error "No %s entry" (if (< count 0) "previous" "next")))))))
 
 (defun devdocs-previous-entry (count)
   "Go backward COUNT entries in this document."
   (interactive "p")
   (devdocs-next-entry (- count)))
 
+(defun devdocs-goto-page (doc page)
+  "Go to a given PAGE (a number or path) of DOC.
+Interactively, read a page name with completion."
+  (interactive (let-alist (car devdocs--stack)
+                 (list .doc (completing-read "Go to page: "
+                                             (append (devdocs--index .doc 'pages) nil)
+                                             nil t nil 'devdocs-history))))
+  (let* ((path (cond ((stringp page) page)
+                     ((numberp page) (elt (devdocs--index doc 'pages) page))))
+         (entry (or (seq-find (lambda (entry) (string= (alist-get 'path entry) path))
+                              (devdocs--index doc 'entries))
+                    `((doc . ,doc) (path . ,path)))))
+    (devdocs--render entry)))
+
+(defun devdocs-first-page (doc)
+  "Go to first page of DOC."
+  (interactive (list (alist-get 'doc (car devdocs--stack))))
+  (devdocs-goto-page doc 0))
+
+(defun devdocs-last-page (doc)
+  "Go to last page of DOC."
+  (interactive (list (alist-get 'doc (car devdocs--stack))))
+  (devdocs-goto-page doc (1- (length (devdocs--index doc 'pages)))))
+
 (defun devdocs-next-page (count)
   "Go forward COUNT pages in this document."
   (interactive "p")
   (let-alist (car devdocs--stack)
-    (let* ((pages (alist-get 'pages (devdocs--index .doc)))
-           (page (+ count (seq-position pages (devdocs--path-file .path))))
-           (path (or (ignore-error 'args-out-of-range (seq-elt pages page))
-                     (user-error (if (< count 0) "No previous page" "No next page")))))
-      (devdocs--render `((doc . ,.doc)
-                         (path . ,path)
-                         (name . ,(format "%s/%s" (1+ page) (length pages))))))))
+    (let* ((pages (devdocs--index .doc 'pages))
+           (dest (+ count (seq-position pages (devdocs--path-file .path)))))
+      (cond ((< dest 0) (user-error "No previous page"))
+            ((<= (length pages) dest) (user-error "No next page")))
+      (devdocs-goto-page .doc dest))))
 
 (defun devdocs-previous-page (count)
   "Go backward COUNT entries in this document."
@@ -349,17 +382,21 @@ with the order of appearance in the text."
       (message "Copied %s" url))))
 
 (let ((map devdocs-mode-map))
-  (define-key map [tab] 'forward-button)
-  (define-key map [backtab] 'backward-button)
-  (define-key map "i" 'devdocs-lookup)
-  (define-key map "p" 'devdocs-previous-entry)
-  (define-key map "n" 'devdocs-next-entry)
-  (define-key map "[" 'devdocs-previous-page)
-  (define-key map "]" 'devdocs-next-page)
-  (define-key map "l" 'devdocs-go-back)
-  (define-key map "r" 'devdocs-go-forward)
-  (define-key map "w" 'devdocs-copy-url)
-  (define-key map "." 'devdocs-goto-target))
+  (define-key map [tab] #'forward-button)
+  (define-key map [backtab] #'backward-button)
+  (define-key map "d" #'devdocs-peruse)
+  (define-key map "i" #'devdocs-lookup)
+  (define-key map "p" #'devdocs-previous-entry)
+  (define-key map "n" #'devdocs-next-entry)
+  (define-key map "g" #'devdocs-goto-page)
+  (define-key map "[" #'devdocs-previous-page)
+  (define-key map "]" #'devdocs-next-page)
+  (define-key map "<" #'devdocs-first-page)
+  (define-key map ">" #'devdocs-last-page)
+  (define-key map "l" #'devdocs-go-back)
+  (define-key map "r" #'devdocs-go-forward)
+  (define-key map "w" #'devdocs-copy-url)
+  (define-key map "." #'devdocs-goto-target))
 
 ;;; Rendering
 
@@ -377,10 +414,9 @@ with the order of appearance in the text."
   (pcase (string-to-char path)
     ('?/ path)
     ('?# (concat (devdocs--path-file base) path))
-    (_ (substring ;; ugly!
+    (_ (seq-rest ;; drop leading slash
         (url-expander-remove-relative-links ;; undocumented function!
-         (concat (file-name-directory base) path))
-        1))))
+         (concat (file-name-directory base) path))))))
 
 (defun devdocs--shr-tag-pre (dom)
   "Insert and fontify pre-tag represented by DOM."
@@ -393,7 +429,7 @@ with the order of appearance in the text."
 (defun devdocs--render (entry)
   "Render a DevDocs documentation entry, returning a buffer.
 
-ENTRY is an alist like those in the variable `devdocs--index',
+ENTRY is an alist like those in the entry index of the document,
 possibly with an additional ENTRY.fragment which overrides the
 fragment part of ENTRY.path."
   (with-current-buffer (get-buffer-create "*devdocs*")
@@ -435,28 +471,35 @@ ARGS is passed as is to `browse-url'."
       (let* ((dest (devdocs--path-expand url .path))
              (file (devdocs--path-file dest))
              (frag (devdocs--path-fragment dest))
-             (entry (seq-some (lambda (it)
-                                (when (let-alist it
-                                        (or (string= .path dest)
-                                            (string= .path file)))
-                                  it))
-                              (alist-get 'entries (devdocs--index .doc)))))
+             (entry (seq-find (lambda (it)
+                                (let-alist it
+                                  (or (string= .path dest)
+                                      (string= .path file))))
+                              (devdocs--index .doc 'entries))))
         (unless entry (error "Can't find `%s'" dest))
-        (push `(doc . ,.doc) entry)
         (when frag (push `(fragment . ,frag) entry))
         (devdocs--render entry)))))
 
-;;; Lookup command
+;;; Lookup commands
 
-(defun devdocs--entries (doc)
-  "A list of entries in DOC, as propertized strings."
-  (seq-map (lambda (it)
-             (let ((s (let-alist it
-                        ;; Disambiguation cookie for entries with same .name
-                        (format #("%s\0%c%s" 2 7 (invisible t))
-                                .name .index .doc.slug))))
-               (prog1 s (put-text-property 0 1 'devdocs--data it s))))
-           (alist-get 'entries (devdocs--index doc))))
+(defun devdocs--entries (documents)
+  "A list of entries in DOCUMENTS, as propertized strings."
+  (let* ((counts (make-hash-table :test 'equal))
+         (mkentry (lambda (it)
+                    (let* ((name (alist-get 'name it))
+                           (count (1+ (gethash name counts 0))))
+                      (puthash name count counts)
+                      `(,name ,count . ,it))))
+         (entries (mapcan (lambda (doc)
+                            (mapcar mkentry
+                                    (devdocs--index doc 'entries)))
+                          documents)))
+    (mapcar (pcase-lambda (`(,name ,count . ,it))
+              (propertize (if (= 1 (gethash name counts))
+                              name
+                            (format devdocs-disambiguated-entry-format name count))
+                          'devdocs--data it))
+            entries)))
 
 (defun devdocs--get-data (str)
   "Get data stored as a string property in STR."
@@ -468,20 +511,10 @@ ARGS is passed as is to `browse-url'."
     (concat " " (propertize " " 'display '(space :align-to 40))
      (devdocs--doc-title .doc) devdocs-separator .type)))
 
-(defun devdocs--eat-cookie (&rest _)
-  "Eat the disambiguation cookie in the minibuffer."
-  (let* ((pos (minibuffer-prompt-end))
-         (max (point-max)))
-    (while (and (< pos max) (/= 0 (char-after pos)))
-      (setq pos (1+ pos)))
-    (when (< pos max)
-      (add-text-properties pos (next-property-change pos nil max)
-                           '(invisible t rear-nonsticky t)))))
-
 (defun devdocs--relevant-docs (ask)
   "Return a list of relevant documents for the current buffer.
-May ask interactively for the desired documents.  If ASK is
-non-nil, ask unconditionally."
+May ask interactively for the desired documents, remembering the
+choice for this buffer.  If ASK is non-nil, ask unconditionally."
   (if ask
       (let ((docs (devdocs--read-document "Documents for this buffer: " t)))
         (prog1 docs
@@ -496,7 +529,7 @@ non-nil, ask unconditionally."
 
 INITIAL-INPUT is passed to `completing-read'"
   (let* ((cands (devdocs--with-cache
-                 (mapcan #'devdocs--entries documents)))
+                 (devdocs--entries documents)))
          (metadata '(metadata
                      (category . devdocs)
                      (annotation-function . devdocs--annotate)))
@@ -504,12 +537,9 @@ INITIAL-INPUT is passed to `completing-read'"
                  (if (eq action 'metadata)
                      metadata
                    (complete-with-action action cands string predicate))))
-         (cand (minibuffer-with-setup-hook
-                   (lambda ()
-                     (add-hook 'after-change-functions 'devdocs--eat-cookie nil t))
-                   (completing-read prompt coll nil t initial-input
-                                    'devdocs-history
-                                    (thing-at-point 'symbol)))))
+         (cand (completing-read prompt coll nil t initial-input
+                                'devdocs-history
+                                (thing-at-point 'symbol))))
     (devdocs--get-data (car (member cand cands)))))
 
 ;;;###autoload
@@ -535,11 +565,7 @@ If INITIAL-INPUT is not nil, insert it into the minibuffer."
 (defun devdocs-peruse (doc)
   "Read a document from the first page."
   (interactive (list (devdocs--read-document "Peruse documentation: ")))
-  (let ((pages (alist-get 'pages (devdocs--index doc))))
-    (pop-to-buffer
-     (devdocs--render `((path . ,(seq-first pages))
-                        (doc . ,doc)
-                        (name . ,(format "%s/%s" 1 (length pages))))))))
+  (pop-to-buffer (devdocs-goto-page doc 0)))
 
 ;;; Compatibility with the old devdocs package
 
